using System;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using System.IO;

namespace T43_Tools
{
    public partial class Form1 : Form
    {
        // Bin memory map
        const UInt32 RAM_START_ADDRESS = 0x0;
        const UInt32 RAM_END_ADDRESS = 0x1ffff;
        const UInt32 RAM_LENGTH = 0x20000;

        const UInt32 BOOT_START_ADDRESS = 0x20000;
        const UInt32 BOOT_END_ADDRESS = 0x2ffff;
        const UInt32 BOOT_LENGTH = 0x10000;

        const UInt32 OS_START_ADDRESS = 0x30000;
        const UInt32 OS_END_ADDRESS = 0x1bffff;
        const UInt32 OS_LENGTH = 0x190000;

        const UInt32 TRANS_START_ADDRESS = 0x1c0000;
        const UInt32 TRANS_END_ADDRESS = 0x1fcb57;
        const UInt32 TRANS_LENGTH = 0x3cb58;
        const UInt32 TRANS_CHKSUM_1_OFFSET = 0;
        const UInt32 TRANS_CHKSUM_2_OFFSET = 0x1c;

        const UInt32 DIAG_START_ADDRESS = 0x1fcb58;
        const UInt32 DIAG_END_ADDRESS = 0x1feb57;
        const UInt32 DIAG_LENGTH = 0x2000;

        const UInt32 SYSTEM_START_ADDRESS = 0x1feb58;
        const UInt32 SYSTEM_END_ADDRESS = 0x1fffff;
        const UInt32 SYSTEM_LENGTH = 0x14a8;

        // Bin file
        byte[] bin;
        byte[] bin_Seg_Ram = new byte[RAM_LENGTH];
        byte[] bin_Seg_Boot = new byte[BOOT_LENGTH];
        byte[] bin_Seg_OS = new byte[OS_LENGTH];
        byte[] bin_Seg_Trans = new byte[TRANS_LENGTH];
        byte[] bin_Seg_Diag = new byte[DIAG_LENGTH];
        byte[] bin_Seg_System = new byte[SYSTEM_LENGTH];

        bool binLoaded = false;

        ushort map_Table_Decimal_Places = 0;

        public Form1()
        {
            InitializeComponent();

            // Disable checksum tab
            tabControl1.TabPages.Remove(TabPage_Checksum);

            // Defaults
            textBox_ParseASM_AddressLength.Text = "6";
        }


        // FUNCTIONS
        ushort[] ParseHexStringToHexArray(string inputString, bool inputDataIs8Bit = false, bool inputDataIsDecimal = false, bool outputDataIs16Bit = true)
        {
            /*  
                Input: Case 1: uint16_t TCM_5672_TCC_Desired_Pressure[] = { 0x0011, 0x0009, 0x0720,
                    0x08E0, 0x09C0, 0x0BE0, 0x0E10, 0x1020 0x12C0, 0x1740
                        };
                    Case 2:
                        0x0011, 0x0009, 0x0720, 0x08E0, 0x09C0, 0x0BE0, 0x0E10, 0x1020
                        0x12C0, 0x1740
                    Case 3:
                        0x00 0x11 0x00 0x09 0x07 0x20 0x08 0xE0 0x09 0xC0 0x0B 0xE0 0x0E 0x10 0x10 0x20
                        0x12 0xC0 0x17 0x40 
                    Case 4: 
                        00 11 00 09 07 20 08 E0 09 C0 0B E0 0E 10 10 20
                        12 C0 17 40
                    Case 5:
                        0011 0009 0720 08E0 09C0 0BE0 0E10 1020
                        12C0 1740
                    Case 6: (Decimal)
                        17 9 1824 2272 2496 3040 3600 4128 4800 5952
                    Case 7: (Decimal with negative values)
                        17 9 1824 2272 2496 3040 3600 4128 -4800 -5952

                Output: Hex 0011 0009 0720 08E0 09C0 0BE0 0E10 1020 12C0 1740 Dec   17    9 1824
                2272 2496 3040 3600 4128 4800 5952

            */

            string[] separators = { ",", " ", "\n", "\r", "\t" };
            IEnumerable<string> hexValues;
            IEnumerable<string> decValues;
            ushort[] values = { 0 };

            try
            {
                // Input data is hex
                if (!inputDataIsDecimal)
                {
                    // For string with 0x prefix; 0x1234 0x21 etc.
                    if (inputString.Contains("0x"))
                    {
                        hexValues = inputString
                            .Split(separators, StringSplitOptions.RemoveEmptyEntries)
                            .Select(s => s.Trim())
                            .Where(s => s.StartsWith("0x", StringComparison.OrdinalIgnoreCase))
                            .Select(s => s.Substring(2));
                    }
                    // For string without 0x prefix; 01 4B etc.
                    else
                    {
                        hexValues = inputString
                                        .Split(separators, StringSplitOptions.RemoveEmptyEntries)
                                        .Select(s => s.Trim());
                    }

                    // An array of our values formatted in hex
                    values = new ushort[hexValues.Count()];
                    values = hexValues.Select(s => (ushort)Convert.ToInt32(s, 16)).ToArray();
                }
                // Input data is decimal
                else
                {
                    decValues = inputString
                        .Split(separators, StringSplitOptions.RemoveEmptyEntries)
                        .Select(s => s.Trim());

                    // An array of our values formatted in hex
                    values = new ushort[decValues.Count()];
                    values = decValues.Select(s => (ushort)Convert.ToInt32(s)).ToArray();
                }

                if (inputDataIsDecimal || outputDataIs16Bit)
                {
                    if (!inputDataIsDecimal && inputDataIs8Bit)
                    {
                        // combine list elements bytes to ushort 
                        List<ushort> ushortList = new List<ushort>();

                        for (int i = 0; i < values.Count(); i += 2)
                        {
                            byte highByte = (byte)values[i];
                            byte lowByte = (byte)values[i + 1];

                            ushortList.Add((ushort)((ushort)(highByte << 8) | lowByte));
                        }
                        return ushortList.ToArray();
                    }
                    else
                    {
                        return values;
                    }
                }
                else // output format = 8 bit
                {
                    if (inputDataIs8Bit)
                    {
                        return values;
                    }
                    else
                    {
                        // split list elements into bytes...
                        List<ushort> list = new List<ushort>();

                        foreach (uint value in values)
                        {
                            byte highByte = (byte)(value >> 8); // Extract high byte
                            byte lowByte = (byte)(value & 0xFF); // Extract low byte

                            list.Add(highByte);
                            list.Add(lowByte);
                        }
                        return list.ToArray();
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show($"Error converting input: {ex.Message}", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            return null;
        }

        static bool ContainsValidHexChars(string input, int length = 8)
        {
            if (input == null)
            {
                return false;
            }

            if (input.Length == length)
            {
                Regex regex = new Regex("[a-f0-9]");
                MatchCollection matches = regex.Matches(input);
                return matches.Count == length;
            }

            return false;
        }

        private string FormatBytesToWord(string input)
        {
            input = input.Trim();
            string output = "";

            // Remove spaces, newlines, and other non-hex characters
            input = Regex.Replace(input, @"\s|[^0-9A-Fa-f,]", "");

            // Split the input string by comma to get individual byte elements
            string[] byteValues = input.Split(',');

            // Convert the byte array to ushort array
            try
            {
                ushort[] ushortArray = new ushort[byteValues.Length];
                for (int i = 0; i < byteValues.Length; i++)
                {
                    byteValues[i] = byteValues[i].Trim(); // Remove leading/trailing spaces
                    ushortArray[i] = Convert.ToUInt16(byteValues[i], 16);
                }

                // Display the ushort array in the output text box as code
                for (int i = 0; i < ushortArray.Length; i += 2)
                {
                    output += $"0x{ushortArray[i]:X2}{ushortArray[i + 1]:X2}";

                    if (i < ushortArray.Length - 2)
                        output += ", ";
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show($"Error converting input: {ex.Message}", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return output;
        }

        void ConvertToDataTable()
        {
            int rows = 0;
            int columns = 0;

            try
            {
                ushort[] rawArray = ParseHexStringToHexArray(textBox_Map_Table_Input.Text, outputDataIs16Bit: checkBox_Map_Table_16Bit_Output.Checked);

                // return if error
                if (rawArray == null)
                {
                    return;
                }

                if (comboBox_Map_Table_AxisHeader.SelectedIndex == 0) // Axis header is present
                {
                    if (checkBox_Map_Table_1D.Checked)
                    {
                        if (checkBox_Map_Table_OrientationIsRow.Checked)
                        {
                            rows = rawArray[0];
                            columns = 1;
                        }
                        else
                        {
                            rows = 1;
                            columns = rawArray[0];
                        }
                        // Check row x column = data length
                        if (rows * columns != rawArray.Length - 1)
                        {
                            throw new Exception("Row x Column != Data Length");
                        }

                        Array.Copy(rawArray, 1, rawArray, 0, rawArray.Length - 1);
                        Array.Resize(ref rawArray, rawArray.Length - 1);
                    }
                    if (checkBox_Map_Table_2D.Checked)
                    {
                        rows = rawArray[1];
                        columns = rawArray[0];
                        // Check both row and column values are > 0
                        if (rows == 0 || columns == 0)
                        {
                            throw new Exception("Row or Column = 0 when 2D table is selected");
                        }
                        // Check row x column = data length
                        if (rows * columns != rawArray.Length - 2)
                        {
                            throw new Exception("Row x Column != Data Length");
                        }
                        // Remove the row and column data from the array
                        Array.Copy(rawArray, 2, rawArray, 0, rawArray.Length - 2);
                        Array.Resize(ref rawArray, rawArray.Length - 2);
                    }

                    // Update the row / column text boxes
                    textBox_Map_Table_Rows.Text = rows.ToString();
                    textBox_Map_Table_Columns.Text = columns.ToString();
                }
                else // Axis header is not present
                {
                    rows = Convert.ToInt16(textBox_Map_Table_Rows.Text);
                    columns = Convert.ToInt16(textBox_Map_Table_Columns.Text);
                    // Check row x column = data length
                    if (rows > 1 && columns > 1)
                    {
                        if (rows * columns != rawArray.Length)
                        {
                            throw new Exception("Row x Column != Data Length");
                        }
                    }
                    else
                    {
                        if (rows * columns != rawArray.Length)
                        {
                            throw new Exception("Row x Column != Data Length");
                        }
                    }
                }

                // Copy the raw array (ushort) to a double array
                double[] dArray = rawArray.Select(i => (double)i).ToArray();

                // Apply the scalar to the numbers in the array
                double[] scaledArray;
                double scalar = Convert.ToDouble(comboBox_Map_Table_Scalar.Text);
                if (checkBox_Map_Table_Multiply.Checked)
                {                    
                    scaledArray = dArray.Select(e => e * scalar).ToArray();
                }
                else // dividing
                {
                    // If signed
                    if (checkBox_Map_Table_Signed.Checked)
                    {
                        scaledArray = dArray.Select(e => e > 0x7fff ? (e - ushort.MaxValue - 1) / scalar : e / scalar).ToArray();
                    }
                    else // Un-signed
                    {
                        scaledArray = dArray.Select(e => e / scalar).ToArray();
                    }
                }

                // Convert the numbers in the array into text
                string outputString = "";
                // Construct the format string
                string format = "0." + new string('0', map_Table_Decimal_Places);

                foreach (double d in scaledArray)
                {
                    if (outputString.Length != 0)
                    {
                        outputString = outputString + ", ";
                    }

                    outputString = outputString + d.ToString(format);
                }

                // Apply formatting to the output string
                outputString = DataTableFormatter(outputString, columns, rows, insertComma: checkBox_Map_Table_Comma.Checked, 
                    insertPadding: checkBox_Map_Table_TextPadding.Checked);
                if (outputString == null)
                {
                    return;
                }

                // Output to the forms text box
                textBox_Map_Table_Output.Text = outputString;
            }
            catch (Exception ex)
            {
                MessageBox.Show($"Error converting input: {ex.Message}", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        string DataTableFormatter(string input, int columns = 0, int rows = 0, int maxElementsPerLine = 8, bool insertComma = true, bool insertPadding = true, bool insertLeadingTab = false)
        {
            string[] elements = input.Split(", ");
            int totalElements = elements.Length;

            try
            {
                if (insertPadding)
                {
                    // Find the length of the longest element
                    int maxLength = elements.Max(e => e.Length);

                    // Remove the event handler to avoid triggering it during text change
                    //textBox_Map_Table_Padding.TextChanged -= new
                    //System.EventHandler(textBox_Map_Table_Padding_TextChanged);

                    //// Update the text
                    //textBox_Map_Table_Padding.Text = maxLength.ToString();

                    //// Re-subscribe to the TextChanged event
                    //textBox_Map_Table_Padding.TextChanged += new
                    //    System.EventHandler(textBox_Map_Table_Padding_TextChanged);


                    // Pad each element with white spaces to match the length of the longest element
                    elements = elements.Select(e => e.PadLeft(maxLength)).ToArray();
                }

                //if (columns > 0 && rows > 0 && columns * rows != totalElements)
                //{

                //    throw new Exception("Row x Column != Data Length");
                //}

                if (columns > 0 && rows > 0)
                {
                    maxElementsPerLine = columns;
                }

                StringBuilder outputBuilder = new StringBuilder();

                for (int i = 0; i < totalElements; i += maxElementsPerLine)
                {
                    int elementsRemaining = Math.Min(maxElementsPerLine, totalElements - i);
                    string[] currentLine = new string[elementsRemaining];
                    Array.Copy(elements, i, currentLine, 0, elementsRemaining);

                    string line;
                    if (insertComma)
                    {
                        line = string.Join(", ", currentLine);
                    }
                    else
                    {
                        line = string.Join(" ", currentLine);
                    }

                    if (insertLeadingTab) { outputBuilder.Append("    "); }// Add 4 spaces at the beginning of each line
                    outputBuilder.Append(line);

                    if (i + maxElementsPerLine < totalElements) // Add , and line break if it's not the last line
                    {
                        if (insertComma) { outputBuilder.Append(","); }
                        outputBuilder.AppendLine();
                    }
                }

                return outputBuilder.ToString();
            }
            catch (Exception ex)
            {
                MessageBox.Show($"Error converting input: {ex.Message}", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            return null;
        }

        string ParseASM(string inputString)
        {
            string outputString = "";
            int firstOperandRowNumber = 0;

            string[] separators = { "\n", "\r" };

            // Convert input string into rows
            string[] rows = inputString.Split(separators, StringSplitOptions.RemoveEmptyEntries);

            // Return if no rows
            if (rows.Length == 0)
            {
                return "";
            }

            //
            // EDIT
            //
            // Trim start of each row
            for (int i = 0; i < rows.Length; i++)
            {
                rows[i] = rows[i].Trim();
            }

            // Remove consecutive spaces
            for (int i = 0; i < rows.Length; i++)
            {
                rows[i] = string.Join(" ", rows[i].Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries)); ;
            }

            // Strip out blanklines
            for (int i = 0; i < rows.Length; i++)
            {
                if (!rows[i].Any(char.IsLetterOrDigit))
                {
                    Array.Copy(rows, i + 1, rows, i, rows.Length - i - 1);
                    Array.Resize(ref rows, rows.Length - 1);
                }
            }

            // Remove header
            for (int i = 0; i < rows.Length; i++)
            {
                if (rows[i].Contains("***") || rows[i].Contains("FUNCTION"))
                {
                    Array.Copy(rows, i + 1, rows, i, rows.Length - i - 1);
                    Array.Resize(ref rows, rows.Length - 1);
                }
            }

            // Remove __stdcall
            for (int i = 0; i < rows.Length; i++)
            {
                if (rows[i].Contains("__stdcall "))
                {
                    rows[i] = rows[i].Replace("__stdcall ", "");
                }
            }

            // Get function name, find the index of the first space then return the rest of the
            // string up until the opening bracket
            string functionName = "";
            int firstSpaceIndex = rows[0].IndexOf(' ');
            if (firstSpaceIndex != -1)
            {
                // Extract characters after the first space
                functionName = rows[0].Substring(firstSpaceIndex + 1);
                functionName = functionName.Remove(functionName.IndexOf("("));
            }

            // Remove xref sub strings
            for (int i = 0; i < rows.Length; i++)
            {
                if (rows[i].IndexOf(" XREF") > 0)
                {
                    rows[i] = rows[i].Remove(rows[i].IndexOf(" XREF"));
                }
                else if (rows[i].IndexOf("(R)") > 0)
                {
                    rows[i] = rows[i].Remove(rows[i].IndexOf("(R)") - 8);
                }
                else if (rows[i].IndexOf(" XREF") > 0)
                {
                    rows[i] = rows[i].Remove(rows[i].IndexOf("(W)") - 8);
                }
            }

            // Strip out blanklines again
            for (int i = 0; i < rows.Length; i++)
            {
                if (!rows[i].Any(char.IsLetterOrDigit))
                {
                    Array.Copy(rows, i + 1, rows, i, rows.Length - i - 1);
                    Array.Resize(ref rows, rows.Length - 1);
                }
            }

            // To detect first opcode row (address block name present - selected / not selected)
                // If selected -     Split each row into columns with ":" and " " as delimiters
                // If not selected - Split each row into columns with " " as delimiters
            // If selected, check...
                // Column 0 ignore
                // If Column 1 == hex digits and width == user input address length
                //    Columns 2, 3 & 4 == hex digits and width == 2
                // Then match found
            // If not selected
                // If Column 0 == hex digits and width == user input address length
                //    Columns 1, 2 & 3 == hex digits and width == 2
                // Then match found
            
            string[,] sArray = new string[rows.Length, 16];
            
            if (checkBox_ParseASM_AddressBlockPresent.Checked)
                separators = new[] { " ", ":" };
            else
                separators = new[] { " " };

            int addressWidth = Convert.ToInt16(textBox_ParseASM_AddressLength.Text);

            // Fill the columns
            for (int i = 0; i < rows.Length; i++)
            {
                string[] rowSplit = rows[i].Split(separators, StringSplitOptions.RemoveEmptyEntries);

                for (int j = 0; j < rowSplit.Length; j++)
                {
                    sArray[i, j] = rowSplit[j];
                }

                // Test if our detect conditions are met
                if (checkBox_ParseASM_AddressBlockPresent.Checked &&
                    ContainsValidHexChars(sArray[i, 1], addressWidth) &&
                        ContainsValidHexChars(sArray[i, 2], 2) &&
                        ContainsValidHexChars(sArray[i, 3], 2) &&
                        ContainsValidHexChars(sArray[i, 4], 2) &&
                        ContainsValidHexChars(sArray[i, 5], 2))
                {
                    firstOperandRowNumber = i; // First opcode found
                    break;
                }
                else if (!checkBox_ParseASM_AddressBlockPresent.Checked &&
                        ContainsValidHexChars(sArray[i, 0], addressWidth) &&
                        ContainsValidHexChars(sArray[i, 1], 2) &&
                        ContainsValidHexChars(sArray[i, 2], 2) &&
                        ContainsValidHexChars(sArray[i, 3], 2) &&
                        ContainsValidHexChars(sArray[i, 4], 2))
                {
                    firstOperandRowNumber = i; // First opcode found
                    break;
                }
            }

            // Strip out the address block name if present
            if (checkBox_ParseASM_AddressBlockPresent.Checked)
            {
                for (int i = firstOperandRowNumber; i < rows.Length; i++)
                {
                    if (rows[i].IndexOf(":") > 0)
                    {
                        rows[i] = rows[i].Remove(0, rows[i].IndexOf(":") + 1);
                    }
                }
            }

            // Remove additional xref function names from above opcode line
            if (firstOperandRowNumber != 0)
            {
                while (rows[firstOperandRowNumber - 1].StartsWith("FUN") || rows[firstOperandRowNumber - 1].Contains(functionName))
                {
                    Array.Copy(rows, firstOperandRowNumber, rows, firstOperandRowNumber - 1, rows.Length - firstOperandRowNumber);
                    Array.Resize(ref rows, rows.Length - 1);
                    firstOperandRowNumber--;
                }
            }

            // Remove any text from below the first opcode line that doesn't start with an opcode or
            // LAB_
            for (int i = firstOperandRowNumber; i < rows.Length; i++)
            {
                if (!(ContainsValidHexChars(rows[i].Substring(0, addressWidth), addressWidth) &&
                      ContainsValidHexChars(rows[i].Substring(addressWidth + 1,  2), 2) &&
                      ContainsValidHexChars(rows[i].Substring(addressWidth + 4,  2), 2) &&
                      ContainsValidHexChars(rows[i].Substring(addressWidth + 7,  2), 2) &&
                      ContainsValidHexChars(rows[i].Substring(addressWidth + 10, 2), 2)))
                {
                    if (!rows[i].StartsWith("LAB_"))
                    {
                        Array.Copy(rows, i + 1, rows, i, rows.Length - i - 1);
                        Array.Resize(ref rows, rows.Length - 1);
                        i--;
                    }
                }
            }

            // Remove any "= ??" substrings
            for (int i = firstOperandRowNumber; i < rows.Length; i++)
            {
                if (rows[i].Contains("= "))
                {
                    rows[i] = rows[i].Remove(rows[i].IndexOf("= "));
                    //i--;
                }
            }

            //
            // BUILD HEADER - Include local parameters
            //
            for (int i = 0; i < firstOperandRowNumber; i++)
            {
                bool stackNewLine = false;
                bool stackNewLineDone = false;

                // Insert new line above stack info
                while (rows[i].Contains("Stack"))
                {
                    if (!stackNewLine)
                    {
                        stackNewLine = true;
                        outputString = outputString + "\r\n";
                    }
                    outputString = outputString + rows[i] + "\r\n";
                    i++;
                }

                // New line after stack info
                if (stackNewLine && !stackNewLineDone)
                {
                    stackNewLineDone = true;
                    outputString = outputString + "\r\n";

                    // Exit if we're at the end
                    if (i == firstOperandRowNumber)
                    {
                        break;
                    }
                }

                // Continue adding local registers & local parameters
                outputString = outputString + rows[i] + "\r\n";

                // Add a new line between registers and locals
                if (rows[i].StartsWith("assume") && !rows[i + 1].StartsWith("assume"))
                {
                    outputString = outputString + "\r\n";
                }

                // Last time here, add new line
                if (firstOperandRowNumber != 1 && i == firstOperandRowNumber - 1)
                {
                    outputString = outputString + "\r\n";
                }

                // Insert additional line to seperate function name from the local parameters
                if (firstOperandRowNumber != 1 && i == 0)
                {
                    outputString = outputString + "\r\n";
                }
            }

            //
            // BUILD ASSEMBLY OPCODES - Includes jump labels
            //
            // Convert rows into columns. Initialize a 2D array to store the columns
            string[,] columns = new string[rows.Length, 16];
            separators = new[] { " " };

            // Fill the columns
            for (int i = 0; i < rows.Length; i++)
            {
                string[] rowSplit = rows[i].Split(separators, StringSplitOptions.RemoveEmptyEntries);

                for (int j = 0; j < rowSplit.Length; j++)
                {
                    columns[i, j] = rowSplit[j];
                }
            }

            // Append the opcodes and operands to the output text string
            for (int i = firstOperandRowNumber; i < columns.GetLength(0); i++)
            {
                if (columns[i, 0].StartsWith("LAB_"))
                {
                    outputString = outputString + "\r\n" + columns[i, 0];
                }

                for (int j = 5; j < columns.GetLength(1); j++)
                {
                    if (columns[i, j] != null)
                    {
                        if (j == 5)
                        {
                            outputString = outputString + columns[i, j].PadRight(8) + " ";
                        }
                        else if (j < 7)
                        {
                            outputString = outputString + columns[i, j] + " ";
                        }
                    }
                }
                outputString = outputString + "\r\n";
            }

            return outputString;
        }

        private void Ghidra_Data_Table_Convert()
        {
            /*
             * Converts a Ghidra data table to a formatted data table Input:
             * TCM_15235_TCC_Apply_Pressure_Ramp              XREF[1]:     FUN_0014f610:0014f790(*),
             * FUN_0014f610:0014f8dc(*),   
             *  001e7e10 00              ??         00h                     ?  ->  00110011 001e7e11
             *  11              ??         11h 001e7e12 00              ??         00h
             *  ?  ->  0011053d 001e7e13 11              ??         11h 001e7e14 05              ??
             *  05h 001e7e15 3d              ??         3Dh    = 001e7e16 04              ??
             *  04h 001e7e17 31              ??         31h    1
             *
             * Output: uint16_t TCM_15235_TCC_Apply_Pressure_Ramp[] = { 0x0011, 0x0011, 0x053D,
             *  0x0431
             *  };
             */

            string dataType = "";
            string tableName = "";
            string dataArray = "";
            string dataTable;

            try
            {
                string input = textBox_Ghidra_Data_Table_Input.Text;

                // Split input into rows
                string[] rows = input.Split(new[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);

                // Retrieve the hex values from each row
                for (int i = 0; i < rows.Length; i++)
                {
                    // Remove leading white spaces
                    rows[i] = rows[i].TrimStart();
                    // Split row into columns
                    string[] s = rows[i].Split(new[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                    // First row and column is the table name
                    if (i == 0)
                    {
                        tableName = s[0];
                    }
                    // Check each column for valid hex values based on the conditions
                    for (int j = 0; j < s.Length; j++)
                    {
                        if (ContainsValidHexChars(s[j], 2))
                        {
                            if (dataArray.Length != 0)
                            {
                                dataArray += ", ";
                            }
                            dataArray += "0x" + s[j].ToUpper();
                        }
                    }
                }

                // Already formatted as 8 bit, just add the datatype label text
                if (checkBox_Ghidra_Data_Table_8Bit.Checked)
                {
                    dataType = "uint8_t ";
                }

                // Convert to 16 bit if selected
                if (checkBox_Ghidra_Data_Table_16Bit.Checked)
                {
                    dataType = "uint16_t ";
                    dataArray = FormatBytesToWord(dataArray);
                }

                dataArray = DataTableFormatter(dataArray, insertLeadingTab: true);

                // Assemble the completed text
                dataTable = dataType + tableName + "[] = {\r\n" + dataArray + "\r\n};";

                // Output to the text box
                textBox_Ghidra_Data_Table_Output.Text = dataTable;
            }
            catch (Exception ex)
            {
                MessageBox.Show($"Error converting input: {ex.Message}", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }


        // GHIDRA DATA TABLE CONVERSION
        private void button_Ghidra_Data_Table_Convert_Click(object sender, EventArgs e)
        {
            Ghidra_Data_Table_Convert();
        }

        private void button_Ghidra_Data_Table_Clear_Click(object sender, EventArgs e)
        {
            textBox_Ghidra_Data_Table_Input.Clear();
            textBox_Ghidra_Data_Table_Output.Clear();
        }

        private void button_Ghidra_Data_Table_Push_Tables_Click(object sender, EventArgs e)
        {
            textBox_Map_Table_Input.Text = textBox_Ghidra_Data_Table_Output.Text;
            textBox_Map_Table_Output.Clear();

            //map_Table_Decimal_Places = 0; textBox_Map_Table_DecimalPlaces.Text =
            //map_Table_Decimal_Places.ToString();

            tabControl1.SelectTab("tabPage2");
        }

        private void checkBox_Ghidra_Data_Table_LittleEndian_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Ghidra_Data_Table_LittleEndian.Checked)
            {
                checkBox_Ghidra_Data_Table_BigEndian.Checked = false;
            }
            else
            {
                checkBox_Ghidra_Data_Table_BigEndian.Checked = true;
            }
        }

        private void checkBox_Ghidra_Data_Table_BigEndian_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Ghidra_Data_Table_BigEndian.Checked)
            {
                checkBox_Ghidra_Data_Table_LittleEndian.Checked = false;
            }
            else
            {
                checkBox_Ghidra_Data_Table_LittleEndian.Checked = true;
            }
        }

        private void checkBox_Ghidra_Data_Table_8Bit_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Ghidra_Data_Table_8Bit.Checked)
            {
                checkBox_Ghidra_Data_Table_16Bit.Checked = false;
            }
            else
            {
                checkBox_Ghidra_Data_Table_16Bit.Checked = true;
            }
            Ghidra_Data_Table_Convert();
        }

        private void checkBox_Ghidra_Data_Table_16Bit_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Ghidra_Data_Table_16Bit.Checked)
            {
                checkBox_Ghidra_Data_Table_8Bit.Checked = false;
            }
            else
            {
                checkBox_Ghidra_Data_Table_8Bit.Checked = true;
            }
            Ghidra_Data_Table_Convert();
        }


        // MAP TABLE GENERATOR
        private void checkBox_Map_Table_8Bit_Output_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_8Bit_Output.Checked)
            {
                checkBox_Map_Table_16Bit_Output.Checked = false;
            }
            else
            {
                checkBox_Map_Table_16Bit_Output.Checked = true;
            }
        }

        private void checkBox_Map_Table_16Bit_Output_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_16Bit_Output.Checked)
            {
                checkBox_Map_Table_8Bit_Output.Checked = false;
            }
            else
            {
                checkBox_Map_Table_8Bit_Output.Checked = true;
            }
        }

        private void checkBox_Map_Table_1D_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_1D.Checked)
            {
                checkBox_Map_Table_2D.Checked = false;

                checkBox_Map_Table_OrientationIsRow.Enabled = true;
                checkBox_Map_Table_OrientationIsColumn.Enabled = true;
            }
            else
            {
                checkBox_Map_Table_2D.Checked = true;
            }
        }

        private void checkBox_Map_Table_2D_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_2D.Checked)
            {
                checkBox_Map_Table_1D.Checked = false;

                checkBox_Map_Table_OrientationIsRow.Enabled = false;
                checkBox_Map_Table_OrientationIsColumn.Enabled = false;
            }
            else
            {
                checkBox_Map_Table_1D.Checked = true;
            }
        }

        private void checkBox_Map_Table_Multiply_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_Multiply.Checked)
            {
                checkBox_Map_Table_Divide.Checked = false;
            }
            else
            {
                checkBox_Map_Table_Divide.Checked = true;
            }

            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void checkBox_Map_Table_Divide_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_Divide.Checked)
            {
                checkBox_Map_Table_Multiply.Checked = false;
            }
            else
            {
                checkBox_Map_Table_Multiply.Checked = true;
            }

            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void checkBox_Map_Table_OrientationIsRow_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_OrientationIsRow.Checked)
            {
                checkBox_Map_Table_OrientationIsColumn.Checked = false;
            }
            else
            {
                checkBox_Map_Table_OrientationIsColumn.Checked = true;
            }
        }

        private void checkBox_Map_Table_OrientationIsColumn_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_OrientationIsColumn.Checked)
            {
                checkBox_Map_Table_OrientationIsRow.Checked = false;
            }
            else
            {
                checkBox_Map_Table_OrientationIsRow.Checked = true;
            }
        }

        private void comboBox_Map_Table_Scalar_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void comboBox_Map_Table_AxisHeader_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (comboBox_Map_Table_AxisHeader.SelectedIndex == 0)
            {
                checkBox_Map_Table_1D.Enabled = true;
                checkBox_Map_Table_2D.Enabled = true;

                checkBox_Map_Table_OrientationIsRow.Enabled = true;
                checkBox_Map_Table_OrientationIsColumn.Enabled = true;

                textBox_Map_Table_Rows.Enabled = false;
                textBox_Map_Table_Columns.Enabled = false;

                label_Table_Rows.Enabled = false;
                label_Table_Columns.Enabled = false;
            }
            else
            {
                checkBox_Map_Table_1D.Enabled = false;
                checkBox_Map_Table_2D.Enabled = false;

                checkBox_Map_Table_OrientationIsRow.Enabled = false;
                checkBox_Map_Table_OrientationIsColumn.Enabled = false;

                textBox_Map_Table_Rows.Enabled = true;
                textBox_Map_Table_Columns.Enabled = true;

                label_Table_Rows.Enabled = true;
                label_Table_Columns.Enabled = true;

                // Prepare initial info
                if (textBox_Map_Table_Rows.Text == "0")
                {
                    textBox_Map_Table_Rows.Text = "1";
                }

                if (textBox_Map_Table_Columns.Text == "0")
                {
                    var count = textBox_Map_Table_Input.Text.Count(x => x == ',');
                    if (count > 0)
                    {
                        count++;
                        textBox_Map_Table_Columns.Text = count.ToString();
                    }
                    else
                    {
                        textBox_Map_Table_Columns.Text = "1";
                    }
                }

            }
        }

        private void button_Map_Table_Clear_Click(object sender, EventArgs e)
        {
            textBox_Map_Table_Input.Clear();
            textBox_Map_Table_Output.Clear();
            textBox_Map_Table_Rows.Clear();
            textBox_Map_Table_Columns.Clear();
        }

        private void button_Map_Table_Convert_Click(object sender, EventArgs e)
        {
            ConvertToDataTable();
        }

        private void textBox_Map_Table_Input_TextChanged(object sender, EventArgs e)
        {

        }

        private void checkBox_Map_Table_Signed_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_Signed.Checked)
            {
                checkBox_Map_Table_UnSigned.Checked = false;
            }
            else
            {
                checkBox_Map_Table_UnSigned.Checked = true;
            }

            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void checkBox_Map_Table_UnSigned_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox_Map_Table_UnSigned.Checked)
            {
                checkBox_Map_Table_Signed.Checked = false;
            }
            else
            {
                checkBox_Map_Table_Signed.Checked = true;
            }

            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void checkBox_Map_Table_Comma_CheckedChanged(object sender, EventArgs e)
        {
            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void checkBox_Map_Table_TextPadding_CheckedChanged(object sender, EventArgs e)
        {
            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void textBox_Map_Table_Padding_TextChanged(object sender, EventArgs e)
        {
            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void button_Map_Table_IncrementDecimalPlaces_Click(object sender, EventArgs e)
        {
            map_Table_Decimal_Places++;
            textBox_Map_Table_DecimalPlaces.Text = map_Table_Decimal_Places.ToString();

            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }

        private void button_Map_Table_DecrementDecimalPlaces_Click(object sender, EventArgs e)
        {
            if (Convert.ToInt16(textBox_Map_Table_DecimalPlaces.Text) > 0)
            {
                map_Table_Decimal_Places--;
                textBox_Map_Table_DecimalPlaces.Text = map_Table_Decimal_Places.ToString();
            }

            if (textBox_Map_Table_Output.Text.Length > 0)
            {
                ConvertToDataTable();
            }
        }


        // PARSE DECOMPILED TEXT
        void button_ParseASM_Convert_Click(object sender, EventArgs e)
        {
            textBox_ParseASM_Output.Clear();
            textBox_ParseASM_Output.Text = ParseASM(textBox_ParseASM_Input.Text);
        }

        void button_ParseASM_Clear_Click(object sender, EventArgs e)
        {
            textBox_ParseASM_Input.Clear();
            textBox_ParseASM_Output.Clear();
        }


        // FORM CONTROL
        private void Form1_Resize(object sender, EventArgs e)
        {
            // Set the TabControl size to fill the form's client area
            tableLayoutPanel1.Size = new Size(this.ClientSize.Width - 296, this.ClientSize.Height - 106);
            tableLayoutPanel2.Size = new Size(this.ClientSize.Width - 296, this.ClientSize.Height - 106);
            tableLayoutPanel4.Size = new Size(this.ClientSize.Width - 296, this.ClientSize.Height - 106);
        }

        // CHECKSUM 
        struct checkSum
        {
            public string name;
            public UInt32 Location_1;
            public UInt32 Location_2;
            public UInt32 StartAddress;
            public UInt32 EndAddress;
            public UInt32 Length;
            public string selectedFileFullPath;
            public string fileName;
        }

        checkSum Checksum_OS;
        checkSum Checksum_Trans;
        checkSum Checksum_Diag;
        checkSum Checksum_System;

        public void btn_Checksum_OpenBin_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.InitialDirectory = "C:\\Users\\Nathan\\OneDrive\\Cars\\VF2 HSV R8 LSA\\GM Programming\\PCMHacking\\T43\\My T43\\Bin Compare Work Folder";
            openFileDialog.Title = "Choose a file";
            openFileDialog.Filter = "BIN Files (*.bin)|*.bin";

            DialogResult result = openFileDialog.ShowDialog();

            if (result == DialogResult.OK)
            {
                string selectedFileFullPath = openFileDialog.FileName;
                string fileName = Path.GetFileName(selectedFileFullPath); // Extract filename

                textBox_Checksum_BinName.Text = fileName;

                bin = File.ReadAllBytes(selectedFileFullPath);

                // Copy from the bin to each segment
                Array.Copy(bin, RAM_START_ADDRESS, bin_Seg_Ram, 0, RAM_LENGTH);
                Array.Copy(bin, BOOT_START_ADDRESS, bin_Seg_Boot, 0, BOOT_LENGTH);
                Array.Copy(bin, OS_START_ADDRESS, bin_Seg_OS, 0, OS_LENGTH);
                Array.Copy(bin, TRANS_START_ADDRESS, bin_Seg_Trans, 0, TRANS_LENGTH);
                Array.Copy(bin, DIAG_START_ADDRESS, bin_Seg_Diag, 0, DIAG_LENGTH);
                Array.Copy(bin, SYSTEM_START_ADDRESS, bin_Seg_System, 0, SYSTEM_LENGTH);

                binLoaded = true;
            }
            else
            {
                binLoaded = false;
            }
        }

        public void tabControl_Checksum_Enter(object sender, EventArgs e)
        {
            Checksum_OS.name = "OS";
            Checksum_Trans.name = "Trans";
            Checksum_Diag.name = "Diag";
            Checksum_System.name = "System";

            TabPage tabPage = sender as TabPage;

            switch (tabPage.Name)
            {
                case "tabControl_Checksum_OS":
                    if (binLoaded == true)
                    {
                        loadChecksumInfo(Checksum_OS, "OS");
                    }
                    break;
                case "tabControl_Checksum_Trans":
                    if (binLoaded == true)
                    {
                        loadChecksumInfo(Checksum_Trans, "Trans");
                    }
                    break;
                case "tabControl_Checksum_Diag":
                    if (binLoaded == true)
                    {
                        loadChecksumInfo(Checksum_Diag, "Diag");
                    }
                    break;
                case "tabControl_Checksum_System":
                    if (binLoaded == true)
                    {
                        loadChecksumInfo(Checksum_System, "System");
                    }
                    break;
            }
        }

        void loadChecksumInfo(checkSum _checkSum, string _name)
        {
            switch (_name)
            {
                case "OS":
                    if (binLoaded == true)
                    {

                    }
                    break;
                case "Trans":
                    if (binLoaded == true)
                    {
                        textBox_Checksum_Trans_SegmentStartAddress.Text = TRANS_START_ADDRESS.ToString("x");
                        textBox_Checksum_Trans_SegmentEndAddress.Text = TRANS_END_ADDRESS.ToString("x");
                        textBox_Checksum_Trans_SegmentLength.Text = TRANS_LENGTH.ToString("x");
                        textBox_Checksum_Trans_Chksum1.Text = bin_Seg_Trans[TRANS_CHKSUM_1_OFFSET].ToString("x") + bin_Seg_Trans[TRANS_CHKSUM_1_OFFSET + 1].ToString("x");
                        textBox_Checksum_Trans_Chksum2.Text = bin_Seg_Trans[TRANS_CHKSUM_2_OFFSET].ToString("x") + bin_Seg_Trans[TRANS_CHKSUM_2_OFFSET + 1].ToString("x");
                        textBox_Checksum_Trans_Chksum1Offset.Text = TRANS_CHKSUM_1_OFFSET.ToString("x");
                        textBox_Checksum_Trans_Chksum2Offset.Text = TRANS_CHKSUM_2_OFFSET.ToString("x");
                    }
                    break;
                case "Diag":
                    if (binLoaded == true)
                    {

                    }
                    break;
                case "System":
                    if (binLoaded == true)
                    {

                    }
                    break;
            }
        }

        private void button_Checksum_Trans_CalcChksum1_Click(object sender, EventArgs e)
        {

        }

        void Calc_Trans_CheckSum_1()
        {
            // If i leave CS2 in the bin i can calc CS1 ok That would mean CS2 is calc'd first, then
            // CS1

            // Looks like i've got it With both CS set to 0; Both CS are MSB (for universal patcher)
            // CS2 = CRC16 with result byte swapped CS1 = Is calc'd after CS1 is written to the bin
            // Word sum segment then 2's compliment 

            byte[] _bytes = new byte[2];
            UInt32 sum = 0;

            for (int i = 0; i < bin_Seg_Trans.Length; i += 2)
            {
                if (i != 0) // Exclude checksum blocks
                {
                    _bytes[0] = bin_Seg_Trans[i];
                    _bytes[1] = bin_Seg_Trans[i + 1];

                    sum += (UInt16)((_bytes[0] << 8) | _bytes[1]);

                }
            }

            sum = ~sum + 1;
            sum = (UInt16)(sum & 0xFFFF);
            textBox_Checksum_Trans_Chksum1_Result.Text = sum.ToString("x");
        }

        // Function to calculate 2's complement of a number represented as an array of two bytes
        byte[] CalculateTwosComplement(byte[] bytes)
        {
            if (bytes.Length != 2)
            {
                throw new ArgumentException("Input array must have exactly two bytes.");
            }

            // Convert bytes to a 16-bit integer
            ushort number = BitConverter.ToUInt16(bytes, 0);

            // Calculate 2's complement
            ushort complement = (ushort)(~number + 1);

            // Convert back to bytes and return
            return BitConverter.GetBytes(complement);
        }
    }
}